@@ -35,10 +35,7 @@
 
 #include <sys/ioctl.h>
 #include <sys/stat.h>
-
-#ifdef HAVE_SYS_FS_H
-# include <linux/fs.h>
-#endif
+#include <linux/fs.h>
 
 #include "nls.h"
 #include "strutils.h"
@@ -49,6 +46,7 @@
 
 #include <libmount.h>
 
+
 #ifndef FITRIM
 struct fstrim_range {
 	uint64_t start;
@@ -58,16 +56,23 @@ struct fstrim_range {
 #define FITRIM		_IOWR('X', 121, struct fstrim_range)
 #endif
 
+struct fstrim_control {
+	struct fstrim_range range;
+
+	unsigned int verbose : 1,
+		     fstab   : 1,
+		     dryrun : 1;
+};
+
 /* returns: 0 = success, 1 = unsupported, < 0 = error */
-static int fstrim_filesystem(const char *path, struct fstrim_range *rangetpl,
-			    int verbose)
+static int fstrim_filesystem(struct fstrim_control *ctl, const char *path, const char *devname)
 {
 	int fd, rc;
 	struct stat sb;
 	struct fstrim_range range;
 
 	/* kernel modifies the range */
-	memcpy(&range, rangetpl, sizeof(range));
+	memcpy(&range, &ctl->range, sizeof(range));
 
 	fd = open(path, O_RDONLY);
 	if (fd < 0) {
@@ -85,6 +90,16 @@ static int fstrim_filesystem(const char *path, struct fstrim_range *rangetpl,
 		rc = -EINVAL;
 		goto done;
 	}
+
+	if (ctl->dryrun) {
+		if (devname)
+			printf(_("%s: 0 B (dry run) trimmed on %s\n"), path, devname);
+		else
+			printf(_("%s: 0 B (dry run) trimmed\n"), path);
+		rc = 0;
+		goto done;
+	}
+
 	errno = 0;
 	if (ioctl(fd, FITRIM, &range)) {
 		rc = errno == EOPNOTSUPP || errno == ENOTTY ? 1 : -errno;
@@ -94,13 +109,19 @@ static int fstrim_filesystem(const char *path, struct fstrim_range *rangetpl,
 		goto done;
 	}
 
-	if (verbose) {
+	if (ctl->verbose) {
 		char *str = size_to_human_string(
 				SIZE_SUFFIX_3LETTER | SIZE_SUFFIX_SPACE,
 				(uint64_t) range.len);
-		/* TRANSLATORS: The standard value here is a very large number. */
-		printf(_("%s: %s (%" PRIu64 " bytes) trimmed\n"),
+		if (devname)
+			/* TRANSLATORS: The standard value here is a very large number. */
+			printf(_("%s: %s (%" PRIu64 " bytes) trimmed on %s\n"),
+				path, str, (uint64_t) range.len, devname);
+		else
+			/* TRANSLATORS: The standard value here is a very large number. */
+			printf(_("%s: %s (%" PRIu64 " bytes) trimmed\n"),
 				path, str, (uint64_t) range.len);
+
 		free(str);
 	}
 
@@ -111,39 +132,55 @@ done:
 	return rc;
 }
 
-static int has_discard(const char *devname, struct sysfs_cxt *wholedisk)
+static int has_discard(const char *devname, struct path_cxt **wholedisk)
 {
-	struct sysfs_cxt cxt, *parent = NULL;
+	struct path_cxt *pc = NULL;
 	uint64_t dg = 0;
 	dev_t disk = 0, dev;
-	int rc;
+	int rc = -1;
 
-	dev = sysfs_devname_to_devno(devname, NULL);
+	dev = sysfs_devname_to_devno(devname);
 	if (!dev)
-		return 1;
+		goto fail;
+
+	pc = ul_new_sysfs_path(dev, NULL, NULL);
+	if (!pc)
+		goto fail;
+
 	/*
 	 * This is tricky to read the info from sys/, because the queue
 	 * attributes are provided for whole devices (disk) only. We're trying
 	 * to reuse the whole-disk sysfs context to optimize this stuff (as
 	 * system usually have just one disk only).
 	 */
-	if (sysfs_devno_to_wholedisk(dev, NULL, 0, &disk) || !disk)
-		return 1;
+	rc = sysfs_blkdev_get_wholedisk(pc, NULL, 0, &disk);
+	if (rc != 0 || !disk)
+		goto fail;
+
 	if (dev != disk) {
-		if (wholedisk->devno != disk) {
-			sysfs_deinit(wholedisk);
-			if (sysfs_init(wholedisk, disk, NULL))
-				return 1;
+		/* Partition, try reuse whole-disk context if valid for the
+		 * current device, otherwise create new context for the
+		 * whole-disk.
+		 */
+		if (*wholedisk && sysfs_blkdev_get_devno(*wholedisk) != disk) {
+			ul_unref_path(*wholedisk);
+			*wholedisk = NULL;
 		}
-		parent = wholedisk;
+		if (!*wholedisk) {
+			*wholedisk = ul_new_sysfs_path(disk, NULL, NULL);
+			if (!*wholedisk)
+				goto fail;
+		}
+		sysfs_blkdev_set_parent(pc, *wholedisk);
 	}
 
-	rc = sysfs_init(&cxt, dev, parent);
-	if (!rc)
-		rc = sysfs_read_u64(&cxt, "queue/discard_granularity", &dg);
+	rc = ul_path_read_u64(pc, &dg, "queue/discard_granularity");
 
-	sysfs_deinit(&cxt);
+	ul_unref_path(pc);
 	return rc == 0 && dg > 0;
+fail:
+	ul_unref_path(pc);
+	return 1;
 }
 
 
@@ -174,23 +211,29 @@ static int uniq_fs_source_cmp(
  * 32 = all failed
  * 64 = some failed, some success
  */
-static int fstrim_all(struct fstrim_range *rangetpl, int verbose)
+static int fstrim_all(struct fstrim_control *ctl)
 {
 	struct libmnt_fs *fs;
 	struct libmnt_iter *itr;
 	struct libmnt_table *tab;
-	struct sysfs_cxt wholedisk = UL_SYSFSCXT_EMPTY;
+	struct libmnt_cache *cache = NULL;
+	struct path_cxt *wholedisk = NULL;
 	int cnt = 0, cnt_err = 0;
+	const char *filename = _PATH_PROC_MOUNTINFO;
 
 	mnt_init_debug(0);
+	ul_path_init_debug();
 
 	itr = mnt_new_iter(MNT_ITER_BACKWARD);
 	if (!itr)
 		err(MNT_EX_FAIL, _("failed to initialize libmount iterator"));
 
-	tab = mnt_new_table_from_file(_PATH_PROC_MOUNTINFO);
+	if (ctl->fstab)
+		filename = mnt_get_fstab_path();
+
+	tab = mnt_new_table_from_file(filename);
 	if (!tab)
-		err(MNT_EX_FAIL, _("failed to parse %s"), _PATH_PROC_MOUNTINFO);
+		err(MNT_EX_FAIL, _("failed to parse %s"), filename);
 
 	/* de-duplicate by mountpoints */
 	mnt_table_uniq_fs(tab, 0, uniq_fs_target_cmp);
@@ -198,15 +241,31 @@ static int fstrim_all(struct fstrim_range *rangetpl, int verbose)
 	/* de-duplicate by source */
 	mnt_table_uniq_fs(tab, MNT_UNIQ_FORWARD, uniq_fs_source_cmp);
 
+	if (ctl->fstab) {
+		cache = mnt_new_cache();
+		if (!cache)
+			err(MNT_EX_FAIL, _("failed to initialize libmount cache"));
+	}
+
 	while (mnt_table_next_fs(tab, itr, &fs) == 0) {
 		const char *src = mnt_fs_get_srcpath(fs),
 			   *tgt = mnt_fs_get_target(fs);
 		char *path;
 		int rc = 1;
 
-		if (!src || !tgt || *src != '/' ||
-		    mnt_fs_is_pseudofs(fs) ||
-		    mnt_fs_is_netfs(fs))
+		if (!tgt || mnt_fs_is_pseudofs(fs) || mnt_fs_is_netfs(fs))
+			continue;
+
+		if (!src && cache) {
+			/* convert LABEL= (etc.) from fstab to paths */
+			const char *spec = mnt_fs_get_source(fs);
+
+			if (!spec)
+				continue;
+			src = mnt_resolve_spec(spec, cache);
+		}
+
+		if (!src || *src != '/')
 			continue;
 
 		/* Is it really accessible mountpoint? Not all mountpoints are
@@ -230,20 +289,21 @@ static int fstrim_all(struct fstrim_range *rangetpl, int verbose)
 		 * This is reason why we ignore EOPNOTSUPP and ENOTTY errors
 		 * from discard ioctl.
 		 */
-		if (fstrim_filesystem(tgt, rangetpl, verbose) < 0)
+		if (fstrim_filesystem(ctl, tgt, src) < 0)
 		       cnt_err++;
 	}
 
-	sysfs_deinit(&wholedisk);
+	ul_unref_path(wholedisk);
 	mnt_unref_table(tab);
 	mnt_free_iter(itr);
+	mnt_unref_cache(cache);
 
 	if (cnt && cnt == cnt_err)
 		return MNT_EX_FAIL;		/* all failed */
 	if (cnt && cnt_err)
 		return MNT_EX_SOMEOK;		/* some ok */
 
-	return EXIT_SUCCESS;
+	return MNT_EX_SUCCESS;
 }
 
 static void __attribute__((__noreturn__)) usage(void)
@@ -257,11 +317,13 @@ static void __attribute__((__noreturn__)) usage(void)
 	fputs(_("Discard unused blocks on a mounted filesystem.\n"), out);
 
 	fputs(USAGE_OPTIONS, out);
-	fputs(_(" -a, --all           trim all mounted filesystems that are supported\n"), out);
+	fputs(_(" -a, --all           trim all supported mounted filesystems\n"), out);
+	fputs(_(" -A, --fstab         trim all supported mounted filesystems from /etc/fstab\n"), out);
 	fputs(_(" -o, --offset <num>  the offset in bytes to start discarding from\n"), out);
 	fputs(_(" -l, --length <num>  the number of bytes to discard\n"), out);
 	fputs(_(" -m, --minimum <num> the minimum extent length to discard\n"), out);
 	fputs(_(" -v, --verbose       print number of discarded bytes\n"), out);
+	fputs(_(" -d, --dry-run       does everything, but trim\n"), out);
 
 	fputs(USAGE_SEPARATOR, out);
 	printf(USAGE_HELP_OPTIONS(21));
@@ -272,17 +334,21 @@ static void __attribute__((__noreturn__)) usage(void)
 int main(int argc, char **argv)
 {
 	char *path = NULL;
-	int c, rc, verbose = 0, all = 0;
-	struct fstrim_range range;
+	int c, rc, all = 0;
+	struct fstrim_control ctl = {
+			.range = { .len = ULLONG_MAX }
+	};
 
 	static const struct option longopts[] = {
 	    { "all",       no_argument,       NULL, 'a' },
+	    { "fstab",     no_argument,       NULL, 'A' },
 	    { "help",      no_argument,       NULL, 'h' },
 	    { "version",   no_argument,       NULL, 'V' },
 	    { "offset",    required_argument, NULL, 'o' },
 	    { "length",    required_argument, NULL, 'l' },
 	    { "minimum",   required_argument, NULL, 'm' },
 	    { "verbose",   no_argument,       NULL, 'v' },
+	    { "dry-run",   no_argument,       NULL, 'n' },
 	    { NULL, 0, NULL, 0 }
 	};
 
@@ -291,14 +357,17 @@ int main(int argc, char **argv)
 	textdomain(PACKAGE);
 	atexit(close_stdout);
 
-	memset(&range, 0, sizeof(range));
-	range.len = ULLONG_MAX;
-
-	while ((c = getopt_long(argc, argv, "ahVo:l:m:v", longopts, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "Aahl:m:no:Vv", longopts, NULL)) != -1) {
 		switch(c) {
+		case 'A':
+			ctl.fstab = 1;
+			/* fallthrough */
 		case 'a':
 			all = 1;
 			break;
+		case 'n':
+			ctl.dryrun = 1;
+			break;
 		case 'h':
 			usage();
 			break;
@@ -306,19 +375,19 @@ int main(int argc, char **argv)
 			printf(UTIL_LINUX_VERSION);
 			return EXIT_SUCCESS;
 		case 'l':
-			range.len = strtosize_or_err(optarg,
+			ctl.range.len = strtosize_or_err(optarg,
 					_("failed to parse length"));
 			break;
 		case 'o':
-			range.start = strtosize_or_err(optarg,
+			ctl.range.start = strtosize_or_err(optarg,
 					_("failed to parse offset"));
 			break;
 		case 'm':
-			range.minlen = strtosize_or_err(optarg,
+			ctl.range.minlen = strtosize_or_err(optarg,
 					_("failed to parse minimum extent length"));
 			break;
 		case 'v':
-			verbose = 1;
+			ctl.verbose = 1;
 			break;
 		default:
 			errtryhelp(EXIT_FAILURE);
@@ -338,14 +407,11 @@ int main(int argc, char **argv)
 	}
 
 	if (all)
-		rc = fstrim_all(&range, verbose);
-	else {
-		rc = fstrim_filesystem(path, &range, verbose);
-		if (rc == 1) {
-			warnx(_("%s: the discard operation is not supported"), path);
-			rc = EXIT_FAILURE;
-		}
-	}
+		return fstrim_all(&ctl);	/* MNT_EX_* codes */
 
-	return rc;
+	rc = fstrim_filesystem(&ctl, path, NULL);
+	if (rc == 1)
+		warnx(_("%s: the discard operation is not supported"), path);
+
+	return rc == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
 }
